/* RCA
** Author: Ryan Turner
** Questions? Email rturner3@nd.edu
** Development Stage: In progress

*/

// Includes necessary header file
#include "RCA.h"
#include <stdio.h>
#include <unistd.h>
#include <sys/param.h>
#include <string>
#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>



using namespace std;

void RCA_usage() {
    printf("Usage: RCA dataset <-b bootstrap> <-p p-value> <-dl delimiter> <-o output>\n");
    exit(1);
}

int print_array(const int* array, int length, char* filename, const char* suffix) {
    FILE *fp = NULL;
    if (filename) {
        char fnm[strlen(filename) + strlen(suffix) + 6];
        strcpy(fnm,filename);
        strcat(fnm,"_");
        strcat(fnm,suffix);
        strcat(fnm,".txt");
        fp = fopen(fnm,"wt");
    }
    
    long int i;
    for (i=0; i<length; i++) {
        if (fp) {
            fprintf(fp,"%d", (int)array[i]);
            if (i!=length-1) { fprintf(fp,"\n"); }
        }
        else {
            printf("%d", (int)array[i]);
            if (i!=length-1) { printf(" "); }
        }
    }
    if (fp) {
        fprintf(fp,"\n");
        fclose(fp);
    }
    else {
        printf("\n");
    }
    
    return 0;
}

int print_matrix(int** matrix, int rows, int cols, char* filename, const char* suffix) {
    FILE *fp = NULL;
    if (filename) {
        char fnm[strlen(filename) + strlen(suffix) + 6];
        strcpy(fnm,filename);
        strcat(fnm,"_");
        strcat(fnm,suffix);
        strcat(fnm,".txt");
        fp = fopen(fnm,"wt");
    }
    
    int i, j;
    for (i=0; i<rows; i++) {
        for(j=0; j<cols; j++) {
            if (fp) {
                if (j) { fprintf(fp,","); }
                fprintf(fp,"%d", (int)matrix[i][j]);
            }
            else {
                if (j) { printf(","); }
                printf("%d", (int)matrix[i][j]);
            }
        }
        if (fp) {
            fprintf(fp,"\n");
        }
        else {
            printf("\n");
        }
    }
    if (fp) {
        fclose(fp);
    }

    return 0;
}




double **parse_data(char *filename, int *numObs, int *numVars, char delimiter) {
    double **dataset = 0;
    
    if(access(filename, F_OK)==-1){
        char cwd[MAXPATHLEN];
        getcwd(cwd, MAXPATHLEN);
        fprintf(stderr, "File %s not found in %s", filename, cwd);
        RCA_usage();
        exit(-2);
    }
    
    std::ifstream ifs;
    ifs.open (filename);
    std::string line;
    std::string sVal;
    double val;
    
    *numObs = 0;
    *numVars = 0;
    while(!ifs.eof() && ! ifs.fail() && !ifs.bad()){
        std::getline(ifs, line );
        if(line.length()>0) {
            if (*(numObs)==0) {
                line.erase(line.find_last_not_of(" \n\r\t")+1);
                if(line.length()>0){
                    std::istringstream lstream(line);
                    while(!lstream.eof()){
                        sVal.clear();
                        std::getline(lstream, sVal, delimiter);
                        val =  atof(sVal.c_str());
                        (*(numVars))++;
                    }
                }
            }
            (*(numObs))++;
        }
    }
    
    ifs.close();
    ifs.open (filename);
    
#ifdef DEBUG
	printf("Number of observations in text file: %d\n", *numObs);
#endif
    
#ifdef DEBUG
	printf("Number of variables in text file: %d\n", *numVars);
#endif
    
    dataset = (double **)malloc((*(numObs)) * sizeof(double));
        
    long i = 0, j = 0;
    int rc = 0;
    
    while(!ifs.eof() && !ifs.fail() && !ifs.bad()){
        std::getline(ifs, line );
        // remove trailing white space
        line.erase(line.find_last_not_of(" \n\r\t")+1);
        if(line.length()>0){
            std::istringstream lstream(line);
            j = 0;
            dataset[i] = (double *)malloc((*(numObs)) * sizeof(double));
            while(!lstream.eof()){
                sVal.clear();
                std::getline(lstream, sVal, ',');
                val =  atof(sVal.c_str());
                dataset[i][j] = val;
                j++;
            }
            i++;
        }
    }
    

    ifs.close();
    return dataset;
}



// Main function set up to interact with command line arguments
int main(int argc, char **argv) {

	int i ,j; // Iterator used to loop through command line arguments
	int bootstrap = 1000; // Default number of bootstraps
	double p_value = 0.05; // Default p-value
	char delimiter = ','; // Default dataset file delimiter
	double *inputmaxValues; // Array of maximum values of variables in dataset
	double *inputminValues; // Array of minimum values of variables in dataset
	int numVars = 0; // Number of variables in the dataset or the number of columns in the dataset matrix
	int numObs = 0; // Number of observations in the dataset or the number of rows in the dataset matrix
	int minmaxnumVars = 0; // Number of variables in the (optional) inputted maximum and minimum file
    char *outFile = NULL;
    
    int *membership = NULL;
    double modularity;
    int **merges = NULL;
    
    
    static const char* c_assignments = "member";
    static const char* c_merges = "merge";
	
	// Confirms that the program is run with the -d argument after the program name
	// Program will not execute properly if not run in this order.
	if(argc == 1) {
		RCA_usage();
	}

	// Loop that deals with all of the inputted command line arguments
	for(i = 2; i<argc; i++) {

		// Only considers arguments that are formatted properly
		if(argv[i][0] == '-') {

			// Deals with all possible proper command line inputs
			switch(argv[i][1]) {

				// -b Bootstrap case
				case 'b':	
					
					// Inputted argument is stored using the atoi function
					bootstrap = atoi(argv[++i]);

					// Checks to make sure the bootstrap value is inputted properly.
					// Resets value to zero if the inputted number is a negative number.
					if(bootstrap<0) {

						printf("Bootstrap input must be a positive number. Default bootstrap value of 0 will be used.\n");
						bootstrap = 0;

					}

					#ifdef DEBUG
					printf("bootstrap = %d\n", bootstrap);
					#endif
					break;

				// Checks for -dl argument, delimiter case
				case 'd':	

					if(argv[i][2] == 'l') {

						// Sets delimiter equal to the character inputted
						delimiter = argv[++i][0];

						// Checks if the input indicates a tab delimiter and sets the variable accordingly
						if ((delimiter == 't') || (delimiter == 'T')) {

							delimiter = '\t';

						}
						
						#ifdef DEBUG
						printf("delimiter = \"%c\"\n", delimiter);
						#endif

					}

					break;				

				// Checks for -o argument, output filename case
				case 'o':

					// Opens output file and checks to make sure it is a valid filename/directory
					//dataOut = fopen(argv[++i], "w");
					outFile = argv[++i];

					break;

				// Checks for -p argument, p-value case
				case 'p':

					// Inputs p-value as a double
					p_value = atof(argv[++i]);

					// Checks that p-value is in the valid range from 0 to 1
					if((p_value>1) || (p_value<0)) {

						printf("p-value is out of range. Valid range for p is: 0<p<1. p will be set to default value of 0.05\n");
						p_value = 0.05;

					}

					#ifdef DEBUG
					printf("p-value = %lf\n", p_value);
					#endif

					break;

				// In the case that an invalid argument is inputted, a warning is printed out
				default:

					printf("Warning: unknown command line argument \"%s\"\n", argv[i]);
					break;

			}

		}

	}
    
    int merge_dim[2];
    double stats[2];
    
    char buf[50];
    
    if(bootstrap) {
        sprintf(buf,"%d bootstraps", bootstrap); 
    }
    else {
        strcpy(buf,"without bootstrapping");
    }
    
    double **dataset = parse_data(argv[1], &numObs, &numVars, delimiter);
    
    printf("Read succesfully %d rows and %d columns\n",numObs,numVars);
    printf("Estimation parameters:\n\t%s\n\tp-value=%.4f\n\n",buf,p_value);
    printf("Running RCA, this may take a while...\n");
    RCA(dataset, numObs, numVars, &membership, &modularity, &merges, merge_dim, stats, bootstrap, p_value);
    
    int group_num = 0;
    
    for (i=0; i<numObs; i++) {
        if(group_num<membership[i]) {
            group_num = membership[i];
        }
    }
    group_num++;
    
    printf("Completed!\n\n");
    printf("Data standardized on: mean=%.4g stdev=%.4g\n",stats[0],stats[1]);
    printf("Dataset partitioned into %d groups\n",group_num);
    printf("Modularity: %.4g\n",modularity);
    if(!outFile) { printf("Class assignments:\n"); }
    print_array(membership, numObs, outFile, c_assignments);
    if(!outFile) { printf("Partitions:\n"); }
    print_matrix(merges,merge_dim[0],merge_dim[1],outFile,c_merges);
    
    free(membership);
    
    for(i=0; i<merge_dim[0]; i++) {
        free(merges[i]);
    }
    free(merges);
    
    return(0);

}



